/*
* 
*	MIDI library package for Tcl 8.2+.  Originally written by Bruce
*	James (bjames@voyager.co.nz); modified by Steffen Traeger
*	(Steffen.Traeger@t-online.de) 
*
*	Copyright (c) 1997-2000 by Bruce James <bjames@voyager.co.nz>
*	Copyright (c) 2001 by Steffen Traeger <Steffen.Traeger@t-online.de>
*
*   MIDI support for Win32:
*
*		midi::getoutdevs
*		midi::getindevs
*	 	midi::openout <device> 
*	 	midi::openin <device> 
*	 	midi::sendshort <status> <data1> <data2>
*		midi::getoutid
*		midi::getinid
*
*	NOTE: midi::openout and midi::openin returns a channel id, where
*   the usual channel commands, read, gets, puts, close can then be used.
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License as published by
*   the Free Software Foundation; either version 2 of the License, or
*   (at your option) any later version.
*
*   This program is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*   GNU General Public License for more details.
*
*   You should have received a copy of the GNU General Public License
*   along with this program; if not, write to the Free Software
*   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*   contact:
*	email:	Steffen.Traeger@t-online.de
*	url:	http://home.t-online.de/home/Steffen.Traeger
*
************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <windows.h>
#include <mmsystem.h>

//#define USE_TCL_STUBS
#include "tcl.h"
#include "tclerrno.h"

EXTERN void		TclWinConvertError _ANSI_ARGS_((DWORD errCode));

/*****************************************************************************
 * Debugging
 *****************************************************************************/

//#define DEBUGLOG 1

#ifdef DEBUGLOG
	#define DebugLog( Message ) DebugLog_( Message )

  	void DebugLog_(
  		char *Message
  		)
  	{
    	Tcl_Channel DebugFile;

    	DebugFile = Tcl_OpenFileChannel( NULL, "c:\\temp\\debug.txt", "a+", 0644);
    	(void) Tcl_Write( DebugFile, Message, strlen(Message) );
    	(void) Tcl_Close( NULL, DebugFile );
  	}
#else
  	#define DebugLog( Message )
#endif


/*****************************************************************************
 * Low level definitions
 *****************************************************************************/

HMIDIOUT handleMidiOut;
HMIDIIN handleMidiIn;

//---------------------------------------------------------------------

#define MIDI_BEGINSYSEX 0xF0

//---------------------------------------------------------------------

#define MIDIHDR_BUFFERSIZE 1000

//---------------------------------------------------------------------

#define SYSEXIN_BUFFERSIZE 3000
char SysExInBuffer[SYSEXIN_BUFFERSIZE] = {0};
int SysExInBufferIndex = 0;


//---------------------------------------------------------------------

typedef struct Rpmidi_MIDIHDR 
{
	HGLOBAL hdrHandle;
	HGLOBAL sysexHandle;
	MIDIHDR* hdrPointer;
} Rpmidi_MIDIHDR;


//---------------------------------------------------------------------

#define MAXBUFFERS 2

typedef struct Rpmidi_ChannelInstance 
{
	union 
	{
  		HMIDIIN   hMidiIn;   // Midi input handle
		HMIDIOUT  hMidiOut;  // Midi output handle
	} hMidi;
	Rpmidi_MIDIHDR rpmidi_MIDIHDR[MAXBUFFERS]; // NB: Two buffers used for input; only one used for output
	BOOL Closing;
	Tcl_Channel tcl_Channel;
	Tcl_ChannelType tcl_ChannelType;
} Rpmidi_ChannelInstance;


/*****************************************************************************
 * Private function prototypes
 *****************************************************************************/

int Rpmidi_InitMIDIHDR ( 
	Rpmidi_MIDIHDR* rpmidi_MIDIHDR, 
	DWORD bufsize 
	);
int Rpmidi_FreeMIDIHDR ( 
	Rpmidi_MIDIHDR* rpmidi_MIDIHDR 
	);
void Rpmidi_Error ( 
	Tcl_Interp *interp,
	UINT Error,
	char* Message
	);
static char *GetMidiErrorMsg(
	UINT err
	) ;

// MIDI Input Routines
void CALLBACK Rpmidi_InFunc(
	HMIDIIN hMidiIn, 
	UINT wMsg, 
	DWORD dwInstance, 
	DWORD dwParam1, 
	DWORD dwParam2);
int Rpmidi_InOpen ( 
	Tcl_Interp *interp,
	UINT wDeviceID
	);
int Rpmidi_InClose (
	ClientData instancePtr,
	Tcl_Interp *interp
	);
int Rpmidi_InInput (
	ClientData instanceData,
	char *buf,
	int bufSize,
	int *errorCodePtr
	);
int Rpmidi_InOutput (
	ClientData instanceData,
	char *buf,
	int toWrite,
	int *errorCodePtr
	);
void Rpmidi_InWatchChannel(
	ClientData instanceData,
	int mask
	);
int Rpmidi_InGetHandleProc (
	ClientData instanceData,
	int direction,
	ClientData *handlePtr
	);

// MIDI Output Routines
void CALLBACK Rpmidi_OutFunc(
	HMIDIOUT hMidiOut, 
	UINT wMsg, 
	DWORD dwInstance, 
	DWORD dwParam1, 
	DWORD dwParam2
	);
int Rpmidi_OutOpen ( 
	Tcl_Interp *interp,
	UINT wDeviceID
	);
int Rpmidi_OutClose (
	ClientData instancePtr,
	Tcl_Interp *interp
	);
int Rpmidi_OutInput (
	ClientData instanceData,
	char *buf,
	int bufSize,
	int *errorCodePtr
	);
int Rpmidi_OutOutput (
	ClientData instanceData,
	char *buf,
	int toWrite,
	int *errorCodePtr
	);
void Rpmidi_OutWatchChannel(
	ClientData instanceData,
	int mask
	);
int Rpmidi_OutGetHandleProc (
	ClientData instanceData,
	int direction,
	ClientData *handlePtr
	);


/*****************************************************************************
 * Exported functions
 *****************************************************************************(

/*****************************************************************************
*
* GetOutDevs_Cmd --
*
* midi::getoutdevs - retrieves the MIDI output devices present
* in the system. 
* (exported)
*
* Arguments:
*	none 	
*
* Returns:
*	list of MIDI output devices or -1 if empty
* 
*/
int GetOutDevs_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{
	MIDIOUTCAPS moc; 
	UINT wDevID;			            // device id of midi in/out port
	UINT wNumDevs;						// number of input or output devices-1
	
	/* retrieves the number of MIDI output devices */
	wNumDevs = midiOutGetNumDevs();

	/* determine product name from device's capabilities */
	for (wDevID = 0; wDevID < wNumDevs; wDevID++) {
	    if (!midiOutGetDevCaps(wDevID, &moc, sizeof(MIDIOUTCAPS))) {
			Tcl_AppendElement(interp, moc.szPname);
		}
	}

	if (wNumDevs = 0) {
		Tcl_AppendElement(interp, "-1");
	}

	return TCL_OK;
}

/*****************************************************************************
*
* GetInDevs_Cmd --
*
* midi::getindevs - retrieves the MIDI input devices present
* in the system. 
* (exported)
*
* Arguments:
*	none 	
*
* Returns:
*	list of MIDI input devices or -1 if empty
* 
*/
int GetInDevs_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{
	MIDIINCAPS mic; 
	UINT wDevID;			            // device id of midi in/out port
	UINT wNumDevs;						// number of input or output devices-1
	
	/* retrieves the number of midi input devices */
	wNumDevs = midiInGetNumDevs();

	/* determine product name from device's capabilities */
	for (wDevID = 0; wDevID < wNumDevs; wDevID++) {
	    if (!midiInGetDevCaps(wDevID, &mic, sizeof(MIDIINCAPS)))	{
			Tcl_AppendElement(interp, mic.szPname);
		}
	}

	if (wNumDevs = 0)
		Tcl_AppendElement(interp, "-1");

	return TCL_OK;
}

/*****************************************************************************
*
* OpenOut_Cmd --
*
* midi::openout - opens a MIDI output device for playback.
* (exported)
*
* Arguments:
*	device -	Identifier of the MIDI output device 	
*
* Returns:
*	leave blank if OK otherwise an error message
* 
*/
int OpenOut_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{

	int iDevID;
	UINT wDevID;
	UINT wNumDevs;

	if (argc > 2) {
		Tcl_SetResult(interp, "wrong # args: midi::openout dev# ", TCL_STATIC);
		return TCL_ERROR;
	}

	/* get args */
	if (Tcl_GetInt(interp, argv[1], &iDevID) != TCL_OK) {
		return TCL_ERROR;
	}

	wDevID = iDevID;
    wNumDevs = midiOutGetNumDevs();


    if ( wDevID >= wNumDevs )
    {
		Tcl_SetResult(interp, "invalid midi output dev#", TCL_STATIC);
  		return TCL_ERROR;
  	}

    return Rpmidi_OutOpen (interp, wDevID);
}

/*****************************************************************************
*
* OpenIn_Cmd --
*
* midi::openin - opens a MIDI input device
* (exported)
*
* Arguments:
*	device -	Identifier of the MIDI input device 	
*
* Returns:
*	channel identifier
* 
*/
int OpenIn_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{

	int iDevID;
	UINT wDevID;
	UINT wNumDevs;

	if (argc > 2) {
		Tcl_SetResult(interp, "wrong # args: midi::openin dev# ", TCL_STATIC);
		return TCL_ERROR;
	}

	/* get args */
	if (Tcl_GetInt(interp, argv[1], &iDevID) != TCL_OK) {
		return TCL_ERROR;
	}

	wDevID = iDevID;
    wNumDevs = midiInGetNumDevs();


    if ( wDevID >= wNumDevs )
    {
		Tcl_SetResult(interp, "invalid midi input dev#", TCL_STATIC);
  	}

    return Rpmidi_InOpen (interp, wDevID);
}

/*****************************************************************************
*
* GetOutID_Cmd --
*
* midi::getoutid - retrieves the device identifier for the current 
* MIDI output device.
* (exported)
*
* Arguments:
*	none 	
*
* Returns:
*	device - Identifier of the current MIDI output device,
*	         an MIDI error essage if there is no open device
* 
*/
int GetOutID_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{

	UINT wDevID;
	UINT err;

	if (!(err = midiOutGetID(handleMidiOut, &wDevID))) {
	    Tcl_SetObjResult(interp, Tcl_NewIntObj((int)wDevID));
	} else {
		Tcl_SetResult(interp, GetMidiErrorMsg(err), TCL_STATIC);
  		return TCL_ERROR;
	}

	return TCL_OK;
}

/*****************************************************************************
*
* GetInID_Cmd --
*
* midi::getinid - retrieves the device identifier for the current 
* MIDI input device.
* (exported)
*
* Arguments:
*	none 	
*
* Returns:
*	device - Identifier of the current MIDI input device,
*	         an MIDI error essage if there is no open device
* 
*/
int GetInID_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{

	UINT wDevID;
	UINT err;

	if (!(err = midiInGetID(handleMidiIn, &wDevID))) {
	    Tcl_SetObjResult(interp, Tcl_NewIntObj((int)wDevID));
	} else {
		Tcl_SetResult(interp, GetMidiErrorMsg(err), TCL_STATIC);
  		return TCL_ERROR;
	}

	return TCL_OK;
}

/*****************************************************************************
*
* SendShort_Cmd --
*
* midi::sendshort - sends a short MIDI message to the specified MIDI output 
* device. Short MIDI messages are three or less byte data messages with 
* a status and two data bytes.
* (exported)
*
* Arguments:
*	bStatus -	status byte
*	bData1  -	first data byte
*	bData2  -	second data byte
*
* Returns:
*	leave blank if OK otherwise an error message
* 
*/
int SendShort_Cmd(
	ClientData clientData, 
	Tcl_Interp *interp,
	int argc,
	char *argv[]
	) 
{

	int bStatus, bData1, bData2;
	DWORD midiMsg;
	UINT err;

	if (argc > 4) {
		Tcl_SetResult(interp, "wrong # args", TCL_STATIC);
		return TCL_ERROR;
	}

	/* get args */
	if (Tcl_GetInt(interp, argv[1], &bStatus) != TCL_OK) {
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &bData1) != TCL_OK) {
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[3], &bData2) != TCL_OK) {
		return TCL_ERROR;
	}

	/* all parameters packed into a DWORD value */
	midiMsg = bStatus | (bData1 << 8) | (bData2 << 16);
 
	if (err = midiOutShortMsg(handleMidiOut, midiMsg)) {
		Tcl_SetResult(interp, GetMidiErrorMsg(err), TCL_STATIC);
		return TCL_ERROR;
	}

	return TCL_OK;
}


/*****************************************************************************
 * Private functions
 *****************************************************************************/

/*****************************************************************************
*
* GetMidiErrorMsg --
*
* Retrieves a textual description for an error identified by the specified
* error code.
* (internal use)
*
* Arguments:
*	err -	error code
*
* Returns:
*	str -	textual error description
*
*/
static char *GetMidiErrorMsg(UINT err) {
	#define BUFFERSIZE 256
	char buffer[BUFFERSIZE];
	char *str = (char *) Tcl_Alloc(256);
	
	if (!(err = midiOutGetErrorText(err, &buffer[0], BUFFERSIZE))) {
		sprintf(str, "%s", &buffer[0]);
	} else {
		sprintf(str, "Bad error number %d!",err);
	} 
	
	return str; 
}

void Rpmidi_Error ( 
	Tcl_Interp *interp,
	UINT Error,
	char* Message
	)
{
	DebugLog( "***RpmidiError\n" );

	Tcl_SetObjResult( interp, Tcl_NewIntObj(Error) );
	// TclWinConvertError(GetLastError());
}


/*************************************************************************************/

int Rpmidi_InitMIDIHDR ( Rpmidi_MIDIHDR* rpmidi_MIDIHDR, DWORD bufsize )
{
	DebugLog(  "Rpmidi_InitMIDIHDR #1\n" );
	rpmidi_MIDIHDR->hdrHandle = GlobalAlloc(GMEM_MOVEABLE + GMEM_SHARE + GMEM_ZEROINIT, sizeof(MIDIHDR));
	rpmidi_MIDIHDR->sysexHandle = GlobalAlloc(GMEM_MOVEABLE + GMEM_SHARE + GMEM_ZEROINIT, bufsize);
	if ( rpmidi_MIDIHDR->hdrHandle == 0 || rpmidi_MIDIHDR->sysexHandle == 0 ) {
	// TclWinConvertError(GetLastError());
    return -1;
  }

	DebugLog(  "Rpmidi_InitMIDIHDR #2\n" );
	rpmidi_MIDIHDR->hdrPointer = GlobalLock(rpmidi_MIDIHDR->hdrHandle);
	// Initialize the MIDIHDR structure
	rpmidi_MIDIHDR->hdrPointer->lpData = GlobalLock(rpmidi_MIDIHDR->sysexHandle);
	rpmidi_MIDIHDR->hdrPointer->dwBufferLength = bufsize;
	rpmidi_MIDIHDR->hdrPointer->dwBytesRecorded = 0;
	rpmidi_MIDIHDR->hdrPointer->dwUser = 0;
	rpmidi_MIDIHDR->hdrPointer->dwFlags = 0;
	rpmidi_MIDIHDR->hdrPointer->lpNext = 0;
	rpmidi_MIDIHDR->hdrPointer->reserved = 0;

	DebugLog(  "Rpmidi_InitMIDIHDR #3\n" );
	return 0;
}


/*************************************************************************************/

int Rpmidi_DeinitMIDIHDR ( Rpmidi_MIDIHDR* rpmidi_MIDIHDR )
{
	DebugLog(  "Rpmidi_DeinitMIDIHDR #1\n" );
	GlobalUnlock(rpmidi_MIDIHDR->sysexHandle);
	GlobalUnlock(rpmidi_MIDIHDR->hdrHandle);
	GlobalFree( rpmidi_MIDIHDR->sysexHandle );
	GlobalFree( rpmidi_MIDIHDR->hdrHandle );

	DebugLog(  "Rpmidi_DeinitMIDIHDR #2\n" );
	return 0;
}

/*****************************************************************************
 * MIDI Input Routines
 *****************************************************************************/

void CALLBACK Rpmidi_InFunc(
	HMIDIIN hMidiIn, 
	UINT wMsg, 
	DWORD dwInstance, 
	DWORD dwParam1, 
	DWORD dwParam2)
{
	Rpmidi_ChannelInstance *instancePtr;
	MIDIHDR *hdrPtr;
	int BufferIndex;
	#ifdef DEBUGLOG
		char Number[30];
	#endif

	switch ( wMsg )
	{
		case MIM_OPEN:
			DebugLog( "Rpmidi_InFunc #1 - MIM_OPEN\n" );
			break;

		case MIM_CLOSE:
			DebugLog( "Rpmidi_InFunc #2 - MIM_CLOSE\n" );
			break;

		case MIM_DATA:
			DebugLog( "Rpmidi_InFunc #3 - MIM_DATA\n" );
			break;

		case MIM_LONGDATA:
			// Handle sysex input messages (MM_MIM_LONGDATA)
			DebugLog( "Rpmidi_InFunc #4 - MIM_LONGDATA\n" );
			instancePtr = (Rpmidi_ChannelInstance *) dwInstance;
			hdrPtr = ((MIDIHDR *) dwParam1);
			for ( BufferIndex = 0; BufferIndex < MAXBUFFERS; BufferIndex++ )
				if ( hdrPtr == instancePtr->rpmidi_MIDIHDR[BufferIndex].hdrPointer )
					break;
			assert ( BufferIndex < MAXBUFFERS );

			#ifdef DEBUGLOG
				DebugLog( "Rpmidi_InFunc - dwBytesRecorded, SysExInBufferIndex\n" );
				ltoa( hdrPtr->dwBytesRecorded, Number, 10 );
				DebugLog( &Number[0] );
				DebugLog( ", " );
				_itoa( SysExInBufferIndex, Number, 10 );
				DebugLog( &Number[0] );
				DebugLog( "\n" );
			#endif
    
			if ( hdrPtr->dwBytesRecorded != 0 )
			{
				DebugLog( "Rpmidi_InFunc #5\n" );
				if ( (DWORD) hdrPtr->dwBytesRecorded + (DWORD) SysExInBufferIndex <= (DWORD) SYSEXIN_BUFFERSIZE ) {
					// Copy the data from the global buffer to the sysex data buffer.
					memcpy( &SysExInBuffer[ SysExInBufferIndex ], hdrPtr->lpData, hdrPtr->dwBytesRecorded );
				SysExInBufferIndex += hdrPtr->dwBytesRecorded;
			}
			else
			{
				DebugLog( "Rpmidi_InFunc #6\n" );
				// Buffer overflow
				// ???
			}
		}

		// If not closing the channel, then prepare for more input
		if ( !instancePtr->Closing )
		{
			DebugLog( "Rpmidi_InFunc #8\n" );
			// Reprepare the buffer and hand it back to the input device
			if ( midiInPrepareHeader( instancePtr->hMidi.hMidiIn, hdrPtr, sizeof(MIDIHDR) ) != 0 )
			{
      			// ?? interp->result = "error preparing buffer";
				// ?? return TCL_ERROR;
			}
			DebugLog( "Rpmidi_InFunc #9\n" );
			if ( midiInAddBuffer( instancePtr->hMidi.hMidiIn, hdrPtr, sizeof(MIDIHDR) ) != 0 )
			{
				DebugLog( "Rpmidi_InFunc #10\n" );
      			// ?? interp->result = "error adding buffer";
				// ?? return TCL_ERROR;
			}
		}
		DebugLog( "Rpmidi_InFunc #11\n" );
		
		break;

		case MIM_ERROR:
			DebugLog( "Rpmidi_InFunc #12 - MIM_ERROR\n" );
			break;

		case MIM_LONGERROR:
			DebugLog( "Rpmidi_InFunc #13 - MIM_LONGERROR\n" );
			break;

		default:
			DebugLog( "Rpmidi_InFunc #14 - default\n" );
			break;
	}
}


//---------------------------------------------------------------------

int Rpmidi_InOpen ( 
	Tcl_Interp *interp,
	UINT wDeviceID
	)
{
	UINT Error;
	char channelName[20];
	int BufferIndex;

	Rpmidi_ChannelInstance *instancePtr = 
    (Rpmidi_ChannelInstance *) Tcl_Alloc( sizeof(Rpmidi_ChannelInstance) );
	instancePtr->Closing = FALSE;
	SysExInBufferIndex = 0;

	DebugLog( "Rpmidi_InOpen #1\n" );
	if ( ( Error = midiInOpen( &(instancePtr->hMidi.hMidiIn), wDeviceID, (DWORD) Rpmidi_InFunc, (DWORD) instancePtr, CALLBACK_FUNCTION) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error opening midi input device" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_InOpen #2\n" );
	for ( BufferIndex = 0; BufferIndex < MAXBUFFERS; BufferIndex++ )
	{
		if ( ( Error = Rpmidi_InitMIDIHDR ( &(instancePtr->rpmidi_MIDIHDR[BufferIndex]), MIDIHDR_BUFFERSIZE ) ) != 0 )
		{
			Rpmidi_Error ( interp, Error, "error allocating buffer" );
			return TCL_ERROR;
		}

		DebugLog( "Rpmidi_InOpen #3\n" );
		if ( ( Error = midiInPrepareHeader( instancePtr->hMidi.hMidiIn, instancePtr->rpmidi_MIDIHDR[BufferIndex].hdrPointer, sizeof(MIDIHDR) ) ) != 0 )
		{
			Rpmidi_Error ( interp, Error, "error preparing buffer" );
			return TCL_ERROR;
		}

		DebugLog( "Rpmidi_InOpen #4\n" );
		if ( ( Error = midiInAddBuffer( instancePtr->hMidi.hMidiIn, instancePtr->rpmidi_MIDIHDR[BufferIndex].hdrPointer, sizeof(MIDIHDR) ) ) != 0 )
		{
			Rpmidi_Error ( interp, Error, "error adding buffer" );
			return TCL_ERROR;
		}
	}

	DebugLog( "Rpmidi_InOpen #5\n" );
	instancePtr->tcl_ChannelType.typeName = "midiin";
	instancePtr->tcl_ChannelType.blockModeProc = NULL;	
	instancePtr->tcl_ChannelType.closeProc = Rpmidi_InClose;
	instancePtr->tcl_ChannelType.inputProc = Rpmidi_InInput;
	instancePtr->tcl_ChannelType.outputProc = Rpmidi_InOutput;
	instancePtr->tcl_ChannelType.seekProc = NULL;
	instancePtr->tcl_ChannelType.setOptionProc = NULL;
	instancePtr->tcl_ChannelType.getOptionProc = NULL;
	instancePtr->tcl_ChannelType.watchProc = Rpmidi_InWatchChannel;
	instancePtr->tcl_ChannelType.getHandleProc = Rpmidi_InGetHandleProc;

	strcpy( channelName, "midiin"  );
	itoa( wDeviceID, &channelName[ strlen( &channelName[0] ) ], 10 );

	instancePtr->tcl_Channel = Tcl_CreateChannel(&(instancePtr->tcl_ChannelType), &channelName[0], (ClientData) instancePtr, TCL_READABLE);

	DebugLog( "Rpmidi_InOpen #6\n" );
	Tcl_RegisterChannel(interp, instancePtr->tcl_Channel);


	DebugLog( "Rpmidi_InOpen #7\n" );
	if ( ( Error = midiInStart( instancePtr->hMidi.hMidiIn ) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error starting midi input" );
		return TCL_ERROR;
	}

	/* set global handle */
	handleMidiIn = instancePtr->hMidi.hMidiIn;

	DebugLog( "Rpmidi_InOpen #8\n" );
	// Return channel name
	strcpy( interp->result, channelName );
	return TCL_OK;
}


//---------------------------------------------------------------------

int Rpmidi_InClose (
	ClientData instanceData,
	Tcl_Interp *interp
	)
{
	UINT Error;
	int BufferIndex;
	int Done;

	Rpmidi_ChannelInstance *instancePtr = (Rpmidi_ChannelInstance *) instanceData;
	instancePtr->Closing = TRUE;

	DebugLog( "Rpmidi_InClose #1\n" );
	if ( ( Error = midiInReset( instancePtr->hMidi.hMidiIn ) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error resetting midi input" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_InClose #2\n" );
	// Loop until header is returned to application
	Done = 0;
	while ( !Done )
	{
		for ( BufferIndex = 0; BufferIndex < MAXBUFFERS; BufferIndex++ )
			if ( (instancePtr->rpmidi_MIDIHDR[BufferIndex].hdrPointer->dwFlags & MHDR_DONE ) == 0 )
				break;
		Done = BufferIndex == MAXBUFFERS;
		if ( !Done )
		{
			SleepEx( (DWORD) 50, (BOOL) TRUE );
		}	
	}

	for ( BufferIndex = 0; BufferIndex < MAXBUFFERS; BufferIndex++ )
	{
		DebugLog( "Rpmidi_InClose #3\n" );
		if ( (Error = midiInUnprepareHeader( instancePtr->hMidi.hMidiIn, instancePtr->rpmidi_MIDIHDR[BufferIndex].hdrPointer, sizeof(MIDIHDR) ) ) != 0 )
		{
			Rpmidi_Error ( interp, Error, "error unpreparing buffers" );
			return TCL_ERROR;
		}

		DebugLog( "Rpmidi_InClose #4\n" );
		if ( ( Error = Rpmidi_DeinitMIDIHDR ( &(instancePtr->rpmidi_MIDIHDR[BufferIndex]) ) ) != 0 )
		{
			Rpmidi_Error ( interp, Error, "error freeing buffers" );
			return TCL_ERROR;
		}
	}

	DebugLog( "Rpmidi_InClose #5\n" );
	Tcl_UnregisterChannel(interp, instancePtr->tcl_Channel);

	DebugLog( "Rpmidi_InClose #6\n" );
	if ( ( Error = midiInClose( instancePtr->hMidi.hMidiIn ) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error closing midi input" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_InClose #7\n" );
	Tcl_Free( (char *) instancePtr );

	DebugLog( "Rpmidi_InClose #8\n" );
	return TCL_OK;
}


//---------------------------------------------------------------------

int Rpmidi_InInput (
	ClientData instanceData,
	char *buf,
	int bufSize,
	int *errorCodePtr
	)
{
	int BytesCopied;

	DebugLog( "Rpmidi_InInput #1\n" );
	if ( SysExInBufferIndex == 0 ) 
	{
		*errorCodePtr = EAGAIN;
		return 0;
	}

	DebugLog( "Rpmidi_InInput #2\n" );
	// Copy available bytes (up to bufsize)
	BytesCopied = min( SysExInBufferIndex, bufSize );
	memcpy( buf, &SysExInBuffer[0], BytesCopied );

	// If still input left, shuffle it all to beginning of buffer
	if ( BytesCopied < SysExInBufferIndex ) 
	{
		memcpy( &SysExInBuffer[0], &SysExInBuffer[BytesCopied], SysExInBufferIndex - BytesCopied );
	}
	SysExInBufferIndex -= BytesCopied;

	DebugLog( "Rpmidi_InInput #3\n" );
	return BytesCopied;
}

//---------------------------------------------------------------------

int Rpmidi_InOutput (
	ClientData instanceData,
	char *buf,
	int toWrite,
	int *errorCodePtr
	)
{
	// Not valid for input device
	*errorCodePtr = EINVAL;
	return -1;
}


//---------------------------------------------------------------------

void Rpmidi_InWatchChannel(
	ClientData instanceData,
	int mask
	)
{
	// nothing to initialise
}


//---------------------------------------------------------------------

int Rpmidi_InGetHandleProc (
	ClientData instanceData,
	int direction,
	ClientData *handlePtr
	)
{
	// Not valid
	return TCL_ERROR;
}


/*****************************************************************************
 * MIDI Output Routines
 *****************************************************************************/

void CALLBACK Rpmidi_OutFunc(
	HMIDIOUT hMidiOut, 
	UINT wMsg, 
	DWORD dwInstance, 
	DWORD dwParam1, 
	DWORD dwParam2
	)
{
  
	// Rpmidi_ChannelInstance *instancePtr;
	// MIDIHDR *hdrPtr;

	switch ( wMsg )
	{
		case MOM_OPEN:
			DebugLog( "Rpmidi_OutFunc #1 - MOM_OPEN\n" );
			break;

		case MOM_CLOSE:
			DebugLog( "Rpmidi_OutFunc #2 - MOM_CLOSE\n" );
			break;

		case MOM_DONE:
			DebugLog( "Rpmidi_OutFunc #3 - MOM_DONE\n" );
			break;

		default:
			DebugLog( "Rpmidi_OutFunc #4 - default\n" );
			break;
	}		
}



//---------------------------------------------------------------------

int Rpmidi_OutOpen ( 
	Tcl_Interp *interp,
	UINT wDeviceID
	)
{
	UINT Error;
	char channelName[20];

	Rpmidi_ChannelInstance *instancePtr = 
    (Rpmidi_ChannelInstance *) Tcl_Alloc( sizeof(Rpmidi_ChannelInstance) );
	instancePtr->Closing = FALSE;


	DebugLog( "Rpmidi_OutOpen #1\n" );
	if ( ( Error = midiOutOpen( &(instancePtr->hMidi.hMidiOut), wDeviceID, (DWORD) Rpmidi_OutFunc, (DWORD) instancePtr, CALLBACK_FUNCTION) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error opening midi output device" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_OutOpen #2\n" );
	instancePtr->tcl_ChannelType.typeName = "midiout";
	instancePtr->tcl_ChannelType.blockModeProc = NULL;	
	instancePtr->tcl_ChannelType.closeProc = Rpmidi_OutClose;
	instancePtr->tcl_ChannelType.inputProc = Rpmidi_OutInput;
	instancePtr->tcl_ChannelType.outputProc = Rpmidi_OutOutput;
	instancePtr->tcl_ChannelType.seekProc = NULL;
	instancePtr->tcl_ChannelType.setOptionProc = NULL;
	instancePtr->tcl_ChannelType.getOptionProc = NULL;
	instancePtr->tcl_ChannelType.watchProc = Rpmidi_OutWatchChannel;
	instancePtr->tcl_ChannelType.getHandleProc = Rpmidi_OutGetHandleProc;


	strcpy( channelName, "midiout"  );
	itoa( wDeviceID, &channelName[ strlen( &channelName[0] ) ], 10 );

	instancePtr->tcl_Channel = 
    Tcl_CreateChannel(&(instancePtr->tcl_ChannelType), &channelName[0], (ClientData) instancePtr, TCL_WRITABLE);

	DebugLog( "Rpmidi_OutOpen #3\n" );
	Tcl_RegisterChannel(interp, instancePtr->tcl_Channel);

	/* set global handle */
	handleMidiOut = instancePtr->hMidi.hMidiOut;

	DebugLog("Rpmidi_OutOpen #4\n");

	// Return channel name
	strcpy( interp->result, channelName );
	return TCL_OK;
}


//---------------------------------------------------------------------

int Rpmidi_OutClose (
	ClientData instanceData,
	Tcl_Interp *interp
	)
{
	UINT Error;

	Rpmidi_ChannelInstance *instancePtr = (Rpmidi_ChannelInstance *) instanceData;
	instancePtr->Closing = TRUE;

	DebugLog( "Rpmidi_OutClose #1\n" );
	if ( ( Error = midiOutReset( instancePtr->hMidi.hMidiOut ) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error resetting midi output" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_OutClose #2\n" );
	Tcl_UnregisterChannel(interp, instancePtr->tcl_Channel);

	DebugLog( "Rpmidi_OutClose #3\n" );
	if ( ( Error = midiOutClose( instancePtr->hMidi.hMidiOut ) ) != 0 )
	{
		Rpmidi_Error ( interp, Error, "error closing midi output" );
		return TCL_ERROR;
	}

	DebugLog( "Rpmidi_OutClose #4\n" );
	Tcl_Free( (char *) instancePtr );

	DebugLog( "Rpmidi_OutClose #5\n" );
	return TCL_OK;
}


//---------------------------------------------------------------------

int Rpmidi_OutInput (
	ClientData instanceData,
	char *buf,
	int bufSize,
	int *errorCodePtr
	)
{
	// Not valid for output device
	*errorCodePtr = EINVAL;
	return -1;
}

//---------------------------------------------------------------------

int Rpmidi_OutOutput (
	ClientData instanceData,
	char *buf,
	int toWrite,
	int *errorCodePtr
	)
{
	UINT Error;
  
	Rpmidi_ChannelInstance *instancePtr = (Rpmidi_ChannelInstance *) instanceData;

	DebugLog( "Rpmidi_OutOutput #1\n" );
	if ( ( Error = Rpmidi_InitMIDIHDR ( &(instancePtr->rpmidi_MIDIHDR[0]), toWrite ) ) != 0 ) 
	{
		TclWinConvertError(GetLastError());
		*errorCodePtr = Tcl_GetErrno();
		return -1;
	}

	DebugLog( "Rpmidi_OutOutput #2\n" );
	if ( midiOutPrepareHeader( instancePtr->hMidi.hMidiOut, instancePtr->rpmidi_MIDIHDR[0].hdrPointer, sizeof(MIDIHDR) ) != 0 )
	{
		TclWinConvertError(GetLastError());
		*errorCodePtr = Tcl_GetErrno();
		return -1;
	}

	// Copy bytes
	memcpy( instancePtr->rpmidi_MIDIHDR[0].hdrPointer->lpData, &buf[0], toWrite );

	DebugLog( "Rpmidi_OutOutput #3\n" );
	// Do the output
	if ( midiOutLongMsg( instancePtr->hMidi.hMidiOut, instancePtr->rpmidi_MIDIHDR[0].hdrPointer, sizeof(MIDIHDR) ) != 0 )
	{
		TclWinConvertError(GetLastError());
		*errorCodePtr = Tcl_GetErrno();
		return -1;
	}

	DebugLog( "Rpmidi_OutOutput #4\n" );
	// Loop until done
	while ( (instancePtr->rpmidi_MIDIHDR[0].hdrPointer->dwFlags & MHDR_DONE ) == 0 )
	{
		SleepEx( (DWORD) 50, (BOOL) TRUE );
	}

	DebugLog( "Rpmidi_OutOutput #5\n" );
	if ( midiOutUnprepareHeader( instancePtr->hMidi.hMidiOut, instancePtr->rpmidi_MIDIHDR[0].hdrPointer, sizeof(MIDIHDR) ) != 0 )
	{
		TclWinConvertError(GetLastError());
		*errorCodePtr = Tcl_GetErrno();
		return -1;
	}

	DebugLog( "Rpmidi_OutOutput #6\n" );
	if ( ( Error = Rpmidi_DeinitMIDIHDR ( &(instancePtr->rpmidi_MIDIHDR[0]) ) ) != 0 ) 
	{
		TclWinConvertError(GetLastError());
		*errorCodePtr = Tcl_GetErrno();
		return -1;
	}

	DebugLog( "Rpmidi_OutOutput #7\n" );
	return toWrite;
}


//---------------------------------------------------------------------

void Rpmidi_OutWatchChannel(
	ClientData instanceData,
	int mask
	)
{
	// nothing to initialise
}


//---------------------------------------------------------------------

int Rpmidi_OutGetHandleProc (
	ClientData instanceData,
	int direction,
	ClientData *handlePtr
	)
{
	// Not valid
	return TCL_ERROR;
}

#ifdef __cplusplus
	extern "C"
#endif
DLLEXPORT int Midi_Init(Tcl_Interp *interp)
{
 
	int code;

	#ifdef USE_TCL_STUBS
		if (Tcl_InitStubs(interp, "8.3", 0) == NULL) {
			return TCL_ERROR;
		}
    #endif

	code = Tcl_PkgProvide(interp, "midi", "0.2");
    if (code != TCL_OK) {
        return code;
    }

    Tcl_CreateCommand(interp, "midi::getoutdevs", GetOutDevs_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::getindevs", GetInDevs_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::openout", OpenOut_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::openin", OpenIn_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::getinid", GetInID_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::getoutid", GetOutID_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "midi::sendshort", SendShort_Cmd, (ClientData) 0, (Tcl_CmdDeleteProc *) NULL);

    return TCL_OK;
}





